/* lexerr.c

   written by Don Maszle
   15 September 1991
   
   Copyright (c) 1993.  Don Maszle, Frederic Bois.  All rights reserved.

   -- Revisions -----
     Logfile:  SCCS/s.lexerr.c
    Revision:  1.1
        Date:  7/14/93
     Modtime:  19:11:12
      Author:  @a
   -- SCCS  ---------
   
   Reports errors and exits program if fatal.
*/

#ifdef __LOCAL_HDR__
#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "malloc.h"

#else
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#endif

#include "hungtype.h"
#include "lex.h"


/* ReportError
   
   Reports error iCode to terminal (one of RE_) and optional
   szMessage.   If iSeverity is set to RE_FATAL, exits program.
*/

void ReportError (PINPUTBUF pibIn, WORD wCode, PSTR szMsg, PSTR szAltMsg)
{
  char cNull = '\0';
  BOOL bFatal   = wCode & RE_FATAL;
  BOOL bWarning = wCode & RE_WARNING;
  
  wCode &= ~(RE_FATAL | RE_WARNING);
  
  if (!szMsg)
    szMsg = &cNull;

  if (wCode) { 
    if (bWarning)
      printf ("* Warning: ");
    else {
      printf ("*** Error: ");
      if (pibIn)
	pibIn->cErrors++;
    }  /* else */
  }  /* if */
  
  if (pibIn)
    if (pibIn->pfileIn || pibIn->iLNPrev)	/* Line number is valid */
      printf ("line %d: ", pibIn->iLineNum);

    else if (wCode != RE_FILENOTFOUND) {	/* Dummy pibIn, show buffer */
      PSTRLEX szTmp;
      szTmp[MAX_LEX-1] = '\0';
      printf ("'%s'...\n  ", strncpy (szTmp, pibIn->pbufOrg, MAX_LEX-1));
    }  /* if */
    
  switch (wCode) {

  case 0:
    break;
      
  default:
    printf ("Unknown error code %x: %s", wCode, szMsg);
  
  case RE_INIT:
    printf ("Initialization error.");
    break;

  case RE_FILENOTFOUND:
    printf ("File not found \"%s\".", szMsg);
    break;
      
  case RE_CANNOTOPEN:
    printf ("Cannot open file \"%s\".", szMsg);
    break;
      
  case RE_UNEXPECTED:
    printf ("Syntax--unexpected character '%c' in input file.", *szMsg);
    break;
      
  case RE_UNEXPNUMBER:
    printf ("Unexpected number %s in input file.", szMsg);
    break;

  case RE_EXPECTED:
    printf ("Expected '%c' before '%c'.", szMsg[1], szMsg[0]);
    break;
	
  case RE_LEXEXPECTED:
    printf ("Expected <%s>", szMsg);
    if (szAltMsg)
      printf (" before '%s'", szAltMsg);
    break;

    /* USER error handling -- Add user error reporting below */

    /* Model generator errors */

  case RE_BADCONTEXT:
    printf ("'%s' used in invalid context.", szMsg);
    break;
	
  case RE_DUPDECL:
    printf ("Duplicate declaration of model variable '%s'.", szMsg);
    break;

  case RE_OUTOFMEM:
    printf ("Out of memory ! : %s", szMsg);
#ifdef __DOS__
    printf ("Free mem: %ld\n", (long) _memavl());
#endif
      break;
      
  case RE_REDEF:
    printf ("'%s' redefined.", szMsg);
    break;

  case RE_EQNTOOLONG:
    printf ("Equation is too long.  Possibly missing terminator.");
    break;

  case RE_BADSTATE:
    printf ("Invalid state identifier '%s'.", szMsg);
    break;
      
  case RE_UNDEFINED:
    printf ("Undefined identifier '%s'.", szMsg);
    break;
       
  case RE_NOINPDEF:
    printf ("Input '%s' is not initialized.", szMsg);
    break;
      
  case RE_NODYNEQN:
    printf ("State variable '%s' has no dynamics.", szMsg);
    break;

  case RE_TOOMANYVARS:
    printf ("Too many %s declarations. Limit is %d.\n",
	    szMsg, *(PINT)szAltMsg);
    break;

	/* Simulation input errors */
    
  case RE_ERRORSINEXP:
    printf ("Errors in definition of experiment %d\n", *(PINT)szMsg);
    break;

  case RE_NOOUTPUTS:
    fprintf (stderr,
	     "Error:  Experiment %d, no outputs specified\n", *(PINT)szMsg);
    break;

  case RE_NOOUTTIMES:
    fprintf (stderr,
          "Error:  Experiment %d, no output times specified\n", *(PINT)szMsg);
    break;

  case RE_INTEGERR:
    fprintf (stderr, "DifSub() returned error %ld: %s\n", *(PLONG) szMsg, szAltMsg);
    break;

  case RE_SPECERR:
    fprintf (stderr, "Errors in specification: %s\n", szMsg);
    break;
    
  case RE_INSUF_POINTS:
    fprintf (stderr, "Insufficient points in file \"%s\"\n", szMsg);
    break;
    
  case RE_DEFOTHER:
    fprintf (stderr, "Already defining other experiment %s\n", szMsg);
    break;
  }  /* switch */

  printf ("\n");
  if (szAltMsg && wCode != RE_LEXEXPECTED)
    printf ("%s\n", szAltMsg);
  
  if (bFatal) {
    printf ("Fatal errors.  Exiting.\n");
    exit (wCode);
  }  /* if */
  
}  /* ReportError */


